"""
Download utilities.
"""

import os
from pathlib import Path

import gdown

from hgd import __weights_version__
from hgd.config import Config

WEIGHTS_URL = "https://github.com/bhky/head-gesture-detection/releases/download/" \
              f"weights_v{__weights_version__}/{Config.weights_filename}"


def _get_home_dir() -> str:
    return str(os.getenv("HGD_HOME", default=Path.home()))


def get_default_weights_path() -> str:
    home_dir = _get_home_dir()
    return os.path.join(home_dir, f".hgd/weights/{Config.weights_filename}")


def download_weights_to(weights_path: str) -> None:
    download_dir = os.path.dirname(os.path.abspath(weights_path))
    os.makedirs(download_dir, exist_ok=True)
    print("Pre-trained weights will be downloaded.")
    gdown.download(WEIGHTS_URL, weights_path)



"""
Model utilities.
"""
import os
from typing import Dict, List, Optional

import numpy as np
import tensorflow as tf
from tensorflow.keras import layers  # pylint: disable=import-error
from tensorflow.keras.models import Model  # pylint: disable=import-error,no-name-in-module

from hgd.config import Config
from hgd._download import download_weights_to, get_default_weights_path


def make_model(weights_path: Optional[str] = get_default_weights_path()) -> Model:
    seq_input = layers.Input(
        shape=(Config.seq_length, Config.num_original_features),
        dtype=tf.float32, name="input"
    )
    # Shape: (batch_size, seq_length, num_features)
    x = seq_input
    x = Preprocessing()(x)
    x = layers.Conv1D(2, 10, strides=2, padding="valid", activation="relu")(x)
    x = layers.Conv1D(1, 5, strides=2, padding="valid", activation="relu")(x)
    x = layers.Flatten()(x)
    x_0 = layers.Dropout(0.01)(x)

    has_motion = layers.Dense(1, activation="sigmoid", name="has_motion")(x_0)
    gesture_probs = layers.Dense(len(Config.gesture_labels),
                                 activation="softmax", name="gesture_probs")(x_0)
    output = layers.Concatenate()([has_motion, gesture_probs])

    model = Model(seq_input, output)

    if weights_path is not None:
        if not os.path.isfile(weights_path):
            download_weights_to(weights_path)
        model.load_weights(weights_path)

    return model


def load_landmarks(npz_path: str) -> Dict[str, List[List[float]]]:
    loaded = np.load(npz_path)
    return {label: loaded[label].tolist() for label in loaded.files}


class Preprocessing(layers.Layer):  # type: ignore

    def __init__(self) -> None:
        super(Preprocessing, self).__init__(name="preprocessing")

    @staticmethod
    def call(
            x: tf.Tensor,
            training: Optional[bool] = None,
            mask: Optional[tf.Tensor] = None
    ) -> tf.Tensor:
        features = x

        # Make landmark features relative to the face box.
        xs_rel = (features[..., 4:10] - features[..., 0:1]) / features[..., 2:3]
        ys_rel = (features[..., 10:] - features[..., 1:2]) / features[..., 3:4]

        # Make stddev features.
        xs_coord_std = tf.math.reduce_std(xs_rel, axis=-1, keepdims=True)
        ys_coord_std = tf.math.reduce_std(ys_rel, axis=-1, keepdims=True)

        # Concatenate new features and exclude unwanted ones.
        features = tf.concat(  # pylint: disable=unexpected-keyword-arg,no-value-for-parameter
            [xs_rel, ys_rel, xs_coord_std, ys_coord_std], axis=-1
        )

        return features



"""
Video utilities.
"""
from typing import List, Optional, Union

import cv2
from mediapipe.python.solutions import drawing_utils as mp_drawing
from mediapipe.python.solutions import face_detection as mp_face

from hgd.config import Config


def video_to_landmarks(
        video_path: Optional[Union[int, str]],
        max_num_frames: Optional[int] = Config.seq_length,
        from_beginning: bool = True,
        end_padding: bool = True
) -> List[List[float]]:
    video_path = video_path if video_path else 0  # For 0, webcam will be used.
    font_scale = 0.6
    line_thickness = 1

    valid_frame_count = 0
    landmarks: List[List[float]] = []
    cap = cv2.VideoCapture(video_path)  # pylint: disable=no-member
    with mp_face.FaceDetection(model_selection=0,
                               min_detection_confidence=0.5) as face_detection:

        while cap.isOpened():
            ret, bgr_frame = cap.read()
            if not ret:
                if video_path == 0:
                    continue  # Ignore empty frame of webcam.
                break  # End of given video.
            if max_num_frames and from_beginning and valid_frame_count >= max_num_frames:
                break

            frame = cv2.cvtColor(bgr_frame, cv2.COLOR_BGR2RGB)  # pylint: disable=no-member
            result = face_detection.process(frame)

            if result and result.detections and len(result.detections) == 1:
                detection = result.detections[0]

                right_eye_rel = mp_face.get_key_point(detection, mp_face.FaceKeyPoint.RIGHT_EYE)
                left_eye_rel = mp_face.get_key_point(detection, mp_face.FaceKeyPoint.LEFT_EYE)
                nose_tip_rel = mp_face.get_key_point(detection, mp_face.FaceKeyPoint.NOSE_TIP)
                mouth_center_rel = mp_face.get_key_point(detection, mp_face.FaceKeyPoint.MOUTH_CENTER)
                right_ear_rel = mp_face.get_key_point(detection, mp_face.FaceKeyPoint.RIGHT_EAR_TRAGION)
                left_ear_rel = mp_face.get_key_point(detection, mp_face.FaceKeyPoint.LEFT_EAR_TRAGION)

                face_box_rel = detection.location_data.relative_bounding_box
                face_box_rel = [
                    max(0.0, face_box_rel.xmin),
                    max(0.0, face_box_rel.ymin),
                    face_box_rel.width,
                    face_box_rel.height,
                ]
                xs = [
                    right_eye_rel.x,
                    left_eye_rel.x,
                    nose_tip_rel.x,
                    mouth_center_rel.x,
                    right_ear_rel.x,
                    left_ear_rel.x,
                ]
                ys = [
                    right_eye_rel.y,
                    left_eye_rel.y,
                    nose_tip_rel.y,
                    mouth_center_rel.y,
                    right_ear_rel.y,
                    left_ear_rel.y,
                ]

                features = [*face_box_rel, *xs, *ys]
                assert len(features) == Config.num_original_features
                landmarks.append(features)

                valid_frame_count += 1
            else:
                detection = None

            if video_path != 0:
                continue

            if detection:
                mp_drawing.draw_detection(frame, detection)

            flipped_frame = cv2.flip(frame, 1)  # pylint: disable=no-member

            text = f"{valid_frame_count} / {max_num_frames}" \
                if max_num_frames else str(valid_frame_count)
            cv2.putText(  # pylint: disable=no-member
                flipped_frame, text, (10, 20),
                cv2.FONT_HERSHEY_SIMPLEX,  # pylint: disable=no-member
                font_scale, (0, 0, 255),
                line_thickness, cv2.LINE_AA  # pylint: disable=no-member
            )

            cv2.imshow(  # pylint: disable=no-member
                "Webcam",
                cv2.cvtColor(flipped_frame, cv2.COLOR_RGB2BGR)  # pylint: disable=no-member
            )

            if cv2.waitKey(1) == ord("q"):  # pylint: disable=no-member
                break

    cap.release()
    cv2.destroyAllWindows()  # pylint: disable=no-member

    if not landmarks:
        return []

    if max_num_frames and not from_beginning:
        landmarks = landmarks[-max_num_frames:]

    if max_num_frames and end_padding and len(landmarks) < max_num_frames:
        last = landmarks[-1]
        landmarks = landmarks + [last] * (max_num_frames - len(landmarks))

    return landmarks


if __name__ == "__main__":
    video_to_landmarks(None, None)



"""
Config.
"""
from dataclasses import dataclass
from typing import Tuple

from hgd import __weights_version__


@dataclass
class Config:
    gesture_labels: Tuple[str, ...] = ("nodding", "turning")
    undefined_gesture_label: str = "undefined"
    stationary_label: str = "stationary"
    npz_filename: str = "landmarks.npz"
    seq_length: int = 60
    num_original_features: int = 16
    weights_filename: str = f"hgd_weights_v{__weights_version__}.h5"



"""
Inference utilities.
"""
from dataclasses import asdict, dataclass
from typing import Any, Dict, Optional, Sequence

import numpy as np
from tensorflow.keras.models import Model  # pylint: disable=import-error,no-name-in-module

from hgd.video import video_to_landmarks
from hgd.config import Config
from hgd.model import make_model


@dataclass
class Result:
    gesture: str
    probabilities: Dict[str, Any]


def postprocess(
        prediction: Sequence[float],
        motion_threshold: float,
        gesture_threshold: float
) -> Dict[str, Any]:
    if not prediction:
        return asdict(Result(Config.undefined_gesture_label, {}))

    gesture_probs = prediction[1:]
    if prediction[0] < motion_threshold:
        label = Config.stationary_label
    else:
        if np.max(gesture_probs) < gesture_threshold:
            label = Config.undefined_gesture_label
        else:
            label = Config.gesture_labels[int(np.argmax(gesture_probs))]
    result = Result(
        gesture=label,
        probabilities={
            "has_motion": prediction[0],
            "gestures": {
                Config.gesture_labels[i]: gesture_probs[i]
                for i in range(len(Config.gesture_labels))
            }
        }
    )
    return asdict(result)


def predict_video(
        video_path: Optional[str] = None,  # None will start a webcam.
        model: Optional[Model] = None,
        max_num_frames: int = Config.seq_length,  # For the pre-trained model.
        from_beginning: bool = True,
        end_padding: bool = True,
        postprocessing: bool = True,
        motion_threshold: float = 0.5,
        gesture_threshold: float = 0.9
) -> Any:
    if model is None:
        model = make_model()
    landmarks = video_to_landmarks(
        video_path, max_num_frames, from_beginning, end_padding
    )

    if landmarks:
        prediction: Sequence[float] = model.predict(
            np.expand_dims(landmarks, axis=0)
        )[0].tolist()
    else:
        prediction = []

    if not postprocessing:
        return prediction
    return postprocess(prediction, motion_threshold, gesture_threshold)


# if __name__ == "__main__":
#     print(predict_video(model=make_model(f"../training/{Config.weights_filename}")))
